@@ -24,7 +24,7 @@ LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 '''
-
+import itertools
 import json
 import inspect
 
@@ -160,7 +160,6 @@ class DjangoDash:
     '''
     #pylint: disable=too-many-instance-attributes
     def __init__(self, name=None, serve_locally=None,
-                 expanded_callbacks=False,
                  add_bootstrap_links=False,
                  suppress_callback_exceptions=False,
                  **kwargs): # pylint: disable=unused-argument, too-many-arguments
@@ -186,7 +185,6 @@ class DjangoDash:
         else:
             self._serve_locally = serve_locally
 
-        self._expanded_callbacks = expanded_callbacks
         self._suppress_callback_exceptions = suppress_callback_exceptions
 
         if add_bootstrap_links:
@@ -268,7 +266,6 @@ class DjangoDash:
             ndid = self._uid
 
         rd = WrappedDash(base_pathname=base_pathname,
-                         expanded_callbacks=self._expanded_callbacks,
                          replacements=replacements,
                          ndid=ndid,
                          serve_locally=self._serve_locally)
@@ -287,26 +284,58 @@ class DjangoDash:
 
         return rd
 
+    @staticmethod
+    def get_expanded_arguments(func, inputs, state):
+        """Analyse a callback function signature to detect the expanded arguments to add when called.
+        It uses the inputs and the state information to identify what arguments are already coming from Dash.
+
+        It returns a list of the expanded parameters to inject (can be [] if nothing should be injected)
+         or None if all parameters should be injected."""
+        n_dash_parameters = len(inputs or []) + len(state or [])
+
+        parameter_types = {kind: [p.name for p in parameters] for kind, parameters in
+                           itertools.groupby(inspect.signature(func).parameters.values(), lambda p: p.kind)}
+        if inspect.Parameter.VAR_KEYWORD in parameter_types:
+            # there is some **kwargs, inject all parameters
+            expanded = None
+        elif inspect.Parameter.VAR_POSITIONAL in parameter_types:
+            # there is a *args, assume all parameters afterwards (KEYWORD_ONLY) are to be injected
+            # some of these parameters may not be expanded arguments but that is ok
+            expanded = parameter_types.get(inspect.Parameter.KEYWORD_ONLY, [])
+        else:
+            # there is no **kwargs, filter argMap to take only the keyword arguments
+            expanded = parameter_types.get(inspect.Parameter.POSITIONAL_OR_KEYWORD, [])[
+                       n_dash_parameters:] + parameter_types.get(inspect.Parameter.KEYWORD_ONLY, [])
+
+        return expanded
+
     def callback(self, output, inputs=None, state=None, events=None):
-        'Form a callback function by wrapping, in the same way as the underlying Dash application would'
-        callback_set = {'output':output,
-                        'inputs':inputs and inputs or dict(),
-                        'state':state and state or dict(),
-                        'events':events and events or dict()}
-        def wrap_func(func, callback_set=callback_set, callback_sets=self._callback_sets): # pylint: disable=dangerous-default-value, missing-docstring
-            callback_sets.append((callback_set, func))
+        '''Form a callback function by wrapping, in the same way as the underlying Dash application would
+        but handling extra arguments provided by dpd.
+
+        It will inspect the signature of the function to ensure only relevant expanded arguments are passed to the callback.
+
+        If the function accepts a **kwargs => all expanded arguments are sent to the function in the kwargs.
+        If the function has a *args => expanded arguments matching parameters after the *args are injected.
+        Otherwise, take all arguments beyond the one provided by Dash (based on the Inputs/States provided).
+
+        '''
+        callback_set = {'output': output,
+                        'inputs': inputs or [],
+                        'state': state or [],
+                        'events': events or []}
+
+        def wrap_func(func):
+            self._callback_sets.append((callback_set, func))
+            # add an expanded attribute to the function with the information to use in dispatch_with_args
+            # to inject properly only the expanded arguments the function can accept
+            # if .expanded is None => inject all
+            # if .expanded is a list => inject only
+            func.expanded = DjangoDash.get_expanded_arguments(func, inputs, state)
             return func
         return wrap_func
 
-    def expanded_callback(self, output, inputs=[], state=[], events=[]): # pylint: disable=dangerous-default-value
-        '''
-        Form an expanded callback.
-
-        This function registers the callback function, and sets an internal flag that mandates that all
-        callbacks are passed the enhanced arguments.
-        '''
-        self._expanded_callbacks = True
-        return self.callback(output, inputs, state, events)
+    expanded_callback = callback
 
     def clientside_callback(self, clientside_function, output, inputs=None, state=None):
         'Form a callback function by wrapping, in the same way as the underlying Dash application would'
@@ -354,12 +383,20 @@ class PseudoFlask(Flask):
     def register_blueprint(self, *args, **kwargs):
         pass
 
+
+def compare(id_python, id_dash):
+    """Compare an id of a dash component as a python object with an id of a component
+    in dash syntax. It handles both id as str or as dict (pattern-matching)"""
+    if isinstance(id_python, dict):
+        return "{" in id_dash and id_python == json.loads(id_dash)
+    return id_python == id_dash
+
+
 class WrappedDash(Dash):
     'Wrapper around the Plotly Dash application instance'
     # pylint: disable=too-many-arguments, too-many-instance-attributes
     def __init__(self,
-                 base_pathname=None, replacements=None, ndid=None,
-                 expanded_callbacks=False, serve_locally=False,
+                 base_pathname=None, replacements=None, ndid=None, serve_locally=False,
                  **kwargs):
 
         self._uid = ndid
@@ -378,7 +415,6 @@ class WrappedDash(Dash):
         self.scripts.config.serve_locally = serve_locally
 
         self._adjust_id = False
-        self._dash_dispatch = not expanded_callbacks
         if replacements:
             self._replacements = replacements
         else:
@@ -387,10 +423,6 @@ class WrappedDash(Dash):
 
         self._return_embedded = False
 
-    def use_dash_dispatch(self):
-        'Indicate if dispatch is using underlying dash code or the wrapped code'
-        return self._dash_dispatch
-
     def use_dash_layout(self):
         '''
         Indicate if the underlying dash layout can be used.
@@ -413,10 +445,16 @@ class WrappedDash(Dash):
         if initial_arguments:
             if isinstance(initial_arguments, str):
                 initial_arguments = json.loads(initial_arguments)
+        else:
+            initial_arguments = {}
+
+        # Define overrides as self._replacements updated with initial_arguments
+        overrides = dict(self._replacements)
+        overrides.update(initial_arguments)
 
         # Walk tree. If at any point we have an element whose id
         # matches, then replace any named values at this level
-        reworked_data = self.walk_tree_and_replace(baseData, initial_arguments)
+        reworked_data = self.walk_tree_and_replace(baseData, overrides)
 
         response_data = json.dumps(reworked_data,
                                    cls=PlotlyJSONEncoder)
@@ -450,10 +488,15 @@ class WrappedDash(Dash):
             replacements = {}
             # look for id entry
             thisID = data.get('id', None)
-            if thisID is not None:
-                replacements = overrides.get(thisID, None) if overrides else None
-                if not replacements:
-                    replacements = self._replacements.get(thisID, {})
+            if isinstance(thisID, dict):
+                # handle case of thisID being a dict (pattern) => linear search in overrides dict
+                for k, v in overrides.items():
+                    if compare(id_python=thisID, id_dash=k):
+                        replacements = v
+                        break
+            elif thisID is not None:
+                # handle standard case of string thisID => key lookup
+                replacements = overrides.get(thisID, {})
             # walk all keys and replace if needed
             for k, v in data.items():
                 r = replacements.get(k, None)
@@ -543,8 +586,6 @@ class WrappedDash(Dash):
 
         if isinstance(output, (list, tuple)):
             fixed_outputs = [self._fix_callback_item(x) for x in output]
-            # Temporary check; can be removed once the library has been extended
-            raise NotImplementedError("django-plotly-dash cannot handle multiple callback outputs at present")
         else:
             fixed_outputs = self._fix_callback_item(output)
 
@@ -589,45 +630,31 @@ class WrappedDash(Dash):
         if len(argMap) > 0:
             argMap['callback_context'] = callback_context
 
-        outputs = []
-        try:
-            if output[:2] == '..' and output[-2:] == '..':
-                # Multiple outputs
-                outputs = output[2:-2].split('...')
-                target_id = output
-                # Special case of a single output
-                if len(outputs) == 1:
-                    target_id = output[2:-2]
-        except:
-            pass
-
-        single_case = False
-        if len(outputs) < 1:
-            try:
-                output_id = output['id']
-                output_property = output['property']
-                target_id = "%s.%s" %(output_id, output_property)
-            except:
-                target_id = output
-                output_id, output_property = output.split(".")
-            single_case = True
-            outputs = [output,]
+        single_case = not(output.startswith('..') and output.endswith('..'))
+        if single_case:
+            # single Output (not in a list)
+            outputs = [output]
+        else:
+            # multiple outputs in a list (the list could contain a single item)
+            outputs = output[2:-2].split('...')
 
         args = []
 
         da = argMap.get('dash_app', None)
 
-        for component_registration in self.callback_map[target_id]['inputs']:
+        callback_info = self.callback_map[output]
+
+        for component_registration in callback_info['inputs']:
             for c in inputs:
-                if c['property'] == component_registration['property'] and c['id'] == component_registration['id']:
+                if c['property'] == component_registration['property'] and compare(id_python=c['id'],id_dash=component_registration['id']):
                     v = c.get('value', None)
                     args.append(v)
                     if da:
                         da.update_current_state(c['id'], c['property'], v)
 
-        for component_registration in self.callback_map[target_id]['state']:
-            for c in state:
-                if c['property'] == component_registration['property'] and c['id'] == component_registration['id']:
+        for component_registration in callback_info['state']:
+            for c in states:
+                if c['property'] == component_registration['property'] and compare(id_python=c['id'],id_dash=component_registration['id']):
                     v = c.get('value', None)
                     args.append(v)
                     if da:
@@ -638,27 +665,32 @@ class WrappedDash(Dash):
         argMap['outputs_list'] = outputs_list
 
         # Special: intercept case of insufficient arguments
-        # This happens when a propery has been updated with a pipe component
+        # This happens when a property has been updated with a pipe component
         # TODO see if this can be attacked from the client end
 
-        if len(args) < len(self.callback_map[target_id]['inputs']):
+        if len(args) < len(callback_info['inputs']):
             return 'EDGECASEEXIT'
 
-        res = self.callback_map[target_id]['callback'](*args, **argMap)
+        callback = callback_info["callback"]
+        # smart injection of parameters if .expanded is defined
+        if callback.expanded is not None:
+            parameters_to_inject = {*callback.expanded, 'outputs_list'}
+            res = callback(*args, **{k: v for k, v in argMap.items() if k in parameters_to_inject})
+        else:
+            res = callback(*args, **argMap)
+
         if da:
-            if single_case and da.have_current_state_entry(output_id, output_property):
-                response = json.loads(res.data.decode('utf-8'))
-                value = response.get('response', {}).get('props', {}).get(output_property, None)
-                da.update_current_state(output_id, output_property, value)
+            root_value = json.loads(res).get('response', {})
 
-            response = json.loads(res)
-            root_value = response.get('response', {})
             for output_item in outputs:
                 if isinstance(output_item, str):
                     output_id, output_property = output_item.split('.')
                     if da.have_current_state_entry(output_id, output_property):
                         value = root_value.get(output_id,{}).get(output_property, None)
                         da.update_current_state(output_id, output_property, value)
+                else:
+                    # todo: implement saving of state for pattern matching ouputs
+                    raise NotImplementedError("Updating state for dict keys (pattern matching) is not yet implemented")
 
         return res
 
